import conf
import logging
log = logging.getLogger(__name__)

SITE_FILENAME='z_moncov.pth'

def disable():
    '''disable coverage collecting'''
    import sys
    sys.settrace(None)


def enable(db=None, dbhost=conf.DBHOST, dbport=conf.DBPORT, dbname=conf.DBNAME,
        whitelist=conf.WHITELIST, blacklist=conf.BLACKLIST):
    '''enable coverage collecting'''
    import collector
    c = collector.Collector(db=db, dbhost=dbhost, dbport=dbport, dbname=dbname,
            whitelist=whitelist, blacklist=blacklist)
    c.start()
    return c


def drop(db=None, host=conf.DBHOST, port=conf.DBPORT, name=conf.DBNAME):
    '''drop the coverage db'''
    try:
        if db is None:
            db = conf.get_db(host, port, name)
    except Exception as e:
        log.warning("couldn't get db %r, %r, %r: %r" % (host, port, name, e.message))
    else:
        try:
            db.connection.drop_database(db)
        except Exception as e:
            log.warning('failed to drop %r: %r' % (db, e.message))
        else:
            log.info("dropped: %r" % db)

def init(db=None, dbhost=conf.DBHOST, dbport=conf.DBPORT, dbname=conf.DBNAME,
        events_count=conf.EVENTS_COUNT, events_totalsize=conf.EVENTS_TOTALSIZE):
    '''initialize necessary events collection'''        # uses 2 collections
    # - lines: indexed by filename,lineno; fields: hits count
    # - events: capped collection for short-term filename,lineno events
    #           storage
    # events is map-reduced to lines with the moncov stats commands
    if db is None:
        db = conf.get_db(dbhost=dbhost, dbport=dbport, dbname=dbname)
    import pymongo
    try:
        db.create_collection(name="events", capped=True, max=events_count,
                            size=events_totalsize)
        pivot = {'event_id': pymongo.helpers.bson.ObjectId()}
        db.last_event.insert(pivot)
    except pymongo.errors.CollectionInvalid as e:
        # already has such collection; ok
        log.info("Events collection already present in %r: %r" % (db, e))
    return db

def get_collecting_code(dbhost, dbport, dbname, whitelist, blacklist):
    '''return the code to use for coverage collecting initialization from a pth file'''
    whitelist = map(lambda regexp: type(regexp) is str and regexp or regexp.pattern, whitelist)
    blacklist = map(lambda regexp: type(regexp) is str and regexp or regexp.pattern, blacklist)
    return 'import moncov; c = moncov.Collector(dbhost=%r, dbport=%r, dbname=%r, whitelist=%r, blacklist=%r); c.start()\n' % \
        (dbhost, dbport, dbname, whitelist, blacklist)

def sys_enable(db=None, dbhost=conf.DBHOST, dbport=conf.DBPORT, dbname=conf.DBNAME, whitelist=conf.WHITELIST, blacklist=conf.BLACKLIST):
    '''enable system-wide coverage stats collecting; requires permissions'''
    if db is not None:
        # convert to connection/port/name
        host, port, name = db.connection.host, db.connection.port, db.name
    code = get_collecting_code(dbhost=dbhost, dbport=dbport, dbname=dbname,
                            whitelist=whitelist, blacklist=blacklist)
    log.debug("code to use: %r" % code)
    import os
    import site
    import re
    # FIXME: implement with setup??
    filename = os.path.join(site.getsitepackages()[1], SITE_FILENAME)
    try:
        with open(filename, 'w+') as fd:
            fd.write(code)
    except Exception as e:
        log.warning("couldn't dump code to %s: %r" % (filename, e))
    else:
        log.info('written: %s' % filename)

def sys_disable(*args, **kvs):
    '''disable system-wide coverage stats collecting'''
    import site
    import os
    filename = os.path.join(site.getsitepackages()[1], SITE_FILENAME)
    # FIXME: implement with setup??
    try:
        os.unlink(filename)
    except Exception as e:
       log.warning("couldn't remove file %s: %r" % (filename, e))
    else:
        log.info('removed: %s' % filename)



'''raw data tracer for Coverage'''
import sys
import threading
import logging
import multiprocessing
log = logging.getLogger(__name__)

_SHOULD_TRACE = {}

def worker(read_end, ready_condition, dbhost, dbport, dbname):
    # the sub-process to dumping the events to mongo
    import threading; threading.settrace(None) # prevent tracing
    import Queue
    import ctl
    import gevent
    from gevent import monkey; monkey.patch_all(thread=True, aggressive=True)


    def tiny_worker(queue, dbhost, dbport, dbname):
        # feed events to mongo
        db = ctl.init(dbhost=dbhost, dbport=dbport, dbname=dbname)
        for events in iter(queue.get, 'STOP'):
            # convert list of event tuples to list of event dictionaries for mongo to process
            db_events = [{'file': filename, 'line': lineno} for filename, lineno in events]
            try:
                db.events.insert(db_events, w=0)
            except Exception as e:
                # FIXME dunno what else to do here
                pass
    
    def stop_workers(queue, workers):
        for worker in workers:
            queue.put('STOP')
        for worker in workers:
            worker.join()
        
    queue = Queue.Queue()
    #workers = [threading.Thread(target=tiny_worker, args=(queue, dbhost, dbport, dbname)) for i in range(10)]
    workers = [gevent.spawn(tiny_worker, queue, dbhost, dbport, dbname) for i in range(1000)]

    for worker in workers:
        worker.start() 
    ready_condition.acquire()
    ready_condition.notify()
    ready_condition.release()

    # process the fat pipe
    while True:
        try:
            events = read_end.recv()
        except EOFError, Exception:
            stop_workers(queue, workers)
            break
        if events == 'STOP':
            stop_workers(queue, workers)
            break
        try:
            queue.put(events)
        except Exception as e:
            # FIXME dunno what else to do here
            pass

class PyTracer(object):
    def __init__(self, dbhost="localhost", dbport=27017, dbname='moncov', blacklist=[], whitelist=[]):
        self.blacklist = blacklist
        self.whitelist = whitelist
        self._should_process = getattr(sys.modules[__name__], '_SHOULD_TRACE')
        # a stack of line lists for caching inserts of executed lines
        self.stack = [[]]
        # uses 2 collections
        # - lines: indexed by filename,lineno; fields: hits count
        # - events: capped collection for short-term filename,lineno events
        #           storage
        # events is map-reduced to lines with the moncov stats commands
        try:
            from multiprocessing import Pipe, Process, Condition
            self.read_end, self.write_end = Pipe()
            ready_condition = Condition()
            self.worker = Process(target=worker, args=(self.read_end, ready_condition,
                                    dbhost, dbport, dbname))
            ready_condition.acquire()
            self.worker.start()
            ready_condition.wait()
            ready_condition.release()
            print 'working'
            

        except Exception as e:
            print >> sys.stderr, "%r, %r error: %r" % (__file__, self, e)
            #self.db = None
            self.enabled = False
            return
        else:
            self.enabled = True
            import atexit
            atexit.register(lambda: self.__del__())
        log.info('%r using: %r' % (self, (dbhost, dbport, dbname)))

    def __del__(self):
        try:
            print 'stopping %r' % self.worker
            self.write_end.send('STOP')
            self.write_end.close()
            self.worker.join(None)
        except:
            pass

    def _trace(self, frame, event, arg_unused):
        """The trace function passed to sys.settrace."""
        # reentrance not allowed
        if not self.enabled: #or not self.db:
            return self._trace

        filename, lineno = frame.f_code.co_filename, frame.f_lineno

        # needs processing?
        if filename not in self._should_process:
            # on black list?
            self.enabled = False # causes a call --- mask out
            self._should_process[filename] = \
                any([regexp.match(filename) for regexp in self.whitelist]) and \
                not any([regexp.match(filename) for regexp in self.blacklist])
            self.enabled = True
        if not self._should_process[filename]:
           return self._trace

        if event == 'call':
            # push new frame line insert list
            self.stack.append([{'file': filename, 'line': lineno}])
            return self._trace

        if event == 'return' or event == 'exception':
            # commit changes for current frame
            if not self.stack:
                return self._trace
            self.enabled = False # causes a call
            lines = self.stack.pop()
            if not lines:
                print >> sys.stderr, "no lines; bailing-out"
                sys.stderr.flush()
                return None
            try:
                #self.db.events.insert(lines, w=0)
                self.write_end.send(lines)
            except Exception as e:
                print >> sys.stderr, "%r, %r error: %r" % (__file__, self, e)
            finally:
                self.enabled = True
            if event == 'exception':
                self.stack.append([])
            return self._trace

        if event != 'line':
            # c_* events not processed
            return self._trace

        # add filename,lineno to processing list
        # self.stack[-1].append({"file": filename, "line": lineno})
        self.stack[-1].append((filename, lineno))
        return self._trace

    def start(self):
        """Start this Tracer."""
        self.enabled = True
        sys.settrace(self._trace)
        return self._trace

    def stop(self):
        """Stop this Tracer."""
        self.enabled = False
        sys.settrace(None)

    def get_stats(self):
        """Return a dictionary of statistics, or None."""
        return None





from tools import GenericMoncovTest
from tools import tracing
from tools import code_submodule_whitelist as whitelist
from tools import tracing_import_code_submodule as tracing_import
from moncov.stats.rate import Rate

class Test(GenericMoncovTest):
    
    def test_01_no_branch(self):
        with tracing_import('no_branch', db=self.db) as module:
            pass
        self.assertResultBranchRate(None)
    
    def test_02_empty_file(self):
        with tracing_import('empty_file', db=self.db) as module:
            pass
        self.assertResultBranchRate(None)        
    
    def test_03_if_true(self):
        with tracing(self.db, whitelist('if_true'), []):
            import code.if_true
        self.assertResultBranchRate(Rate(1, 1))

    def test_04_if_false(self):
        with tracing(self.db, whitelist('if_false'), []):
            import code.if_false
        self.assertResultBranchRate(Rate(0))

    def test_05_if_true_else(self):
        with tracing(self.db, whitelist('if_true_else'), []):
            import code.if_true_else
        self.assertResultBranchRate(Rate(1, 2))

    def test_06_if_false_else(self):
        with tracing(self.db, whitelist('if_false_else'), []):
            import code.if_false_else
        self.assertResultBranchRate(Rate(1, 2))    

    def test_07_if_true_if_true(self):
        with tracing(self.db, whitelist('if_true_if_true'), []):
            import code.if_true_if_true
        self.assertResultBranchRate(Rate(2, 2))        
        
    def test_08_if_true_if_false(self):
        with tracing_import('if_true_if_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 4))        
        
    def test_09_if_false_if_true_else(self):
        with tracing_import('if_false_if_true_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 3))

    def test_10_if_true_if_true_else(self):
        with tracing_import('if_true_if_true_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))  
        
    def test_11_if_false_for_some(self):
        with tracing_import('if_false_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 3))
        
    def test_12_if_true_for_some_ifelse(self):
        with tracing_import('if_true_for_some_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))

    def test_13_if_true_for_some_ifelse_for_some(self):
        with tracing_import('if_true_for_some_ifelse_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 4))

    def test_14_if_false_for_some_ifelse_for_some(self):
        with tracing_import('if_false_for_some_ifelse_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 4))
        
    def test_15_if_false_while_some(self):
        with tracing_import('if_false_while_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 2))     
        
    def test_16_if_true_while_some_ifelse(self):
        with tracing_import('if_true_while_some_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))
        
    def test_17_if_true_while_false_ifelse(self):
        with tracing_import('if_true_while_false_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 3))    
   
    def test_18_if_false_while_false_if_true_while_false(self):
        with tracing_import('if_false_while_false_if_true_while_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 4)) 
        
    def test_19_while_some(self):
        with tracing_import('while_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 1))        

    def test_20_while_false(self):
        with tracing_import('while_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 1))          

    def test_21_while_some_if_true(self):
        with tracing_import('while_some_if_true', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 2))
        
    def test_22_while_some_if_false(self):
        with tracing_import('while_some_if_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))     
        
    def test_23_while_false_if_true_else(self):
        with tracing_import('while_false_if_true_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 3))
        
    def test_24_while_some_for_some(self):
        with tracing_import('while_some_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 2))
        
    def test_25_while_false_for_some(self):
        with tracing_import('while_false_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 3))
        
    def test_26_while_false_while_some(self):
        with tracing_import('while_false_while_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 2))
        
    def test_27_while_false_while_false(self):
        with tracing_import('while_false_while_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 2))
        
    def test_28_while_some_while_some(self):
        with tracing_import('while_some_while_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 2))
        
    def test_29_while_some_while_false(self):
        with tracing_import('while_some_while_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))

    def test_30_while_false_if_false_ifelse(self):
        with tracing_import('while_false_if_false_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 3))     

    def test_31_for_empty(self):
        with tracing_import('for_empty', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(0, 1))

    def test_32_for_some(self):
        with tracing_import('for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 1))

    def test_33_for_some_for_some(self):
        with tracing_import('for_some_for_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 2))
        
    def test_34_for_some_if_true_ifelse(self):
        with tracing_import('for_some_if_true_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))
        
    def test_35_for_some_if_false_ifelse(self):
        with tracing_import('for_some_if_false_ifelse', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))
        
    def test_36_for_some_while_some(self):
        with tracing_import('for_some_while_some', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 2))
        
    def test_37_for_some_while_false(self):
        with tracing_import('for_some_while_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))
    
    def test_38_for_some_break_else(self):
        with tracing_import('for_some_break_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))

    def test_39_for_empty_pass_else(self):
        with tracing_import('for_empty_pass_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))

    def test_40_for_empty_break_else(self):
        with tracing_import('for_empty_break_else', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 2))
        
    def test_41_for_empty_if_true_if_false_not_nested(self):
        with tracing_import('for_empty_if_true_if_false_not_nested', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(1, 3))

    def test_42_true_true_true_false__function_a_b_if_a_if_b_else(self):
        with tracing_import('function_a_b_if_a_if_b_else', db=self.db) as module:
            module.function(True, True); module.function(True, False)
        self.assertResultBranchRate(Rate(3, 3))

    def test_43_true_fucntion_value_if_value_else(self):
        with tracing_import('function_value_if_value_else', db=self.db) as module: 
            module.function(True)
        self.assertResultBranchRate(Rate(1, 2))

    def test_44_false_fucntion_value_if_value_else(self):
        with tracing_import('function_value_if_value_else', db=self.db) as module: 
            module.function(False)
        self.assertResultBranchRate(Rate(1, 2))

    def test_45_true_false__function_value_if_value_else(self):
        with tracing_import('function_value_if_value_else', db=self.db) as module: 
            module.function(True); module.function(False)
        self.assertResultBranchRate(Rate(2, 2))

    def test_46_if_true_def_if_true_else(self):
        with tracing_import('if_true_def_if_true_false', db=self.db) as module:
            pass
        self.assertResultBranchRate(Rate(2, 3))



import yaml
import logging
import re
import pymongo

log = logging.getLogger(__name__)

try:
    with open('/etc/moncov.yaml') as fd:
        CFG = yaml.load(fd.read())
except Exception as e:
    log.warning("can't read config: %s" % e.message)
    CFG = {}

DBHOST = CFG.get('dbhost', 'localhost')
DBPORT = CFG.get('dbport', 27017)
DBNAME = CFG.get('dbname', 'moncov')
EVENTS_COUNT = CFG.get('events_count', 8192)
EVENTS_TOTALSIZE = CFG.get('events_totalsize', EVENTS_COUNT * 1024)

try:
    WHITELIST = map(lambda item: re.compile(item), CFG.get('whitelist', ['.*']))
except Exception as e:
    log.warning("can't process whitelist, using: ['.*']; %s" % e.message)
    WHITELIST = [re.compile('.*')]
try:
    BLACKLIST = map(lambda item: re.compile(item), CFG.get('blacklist', []))
except Exception as e:
    log.warning("can't process blacklist, using: []; %s" % e.message)
    BLACKLIST = []

def get_connection(dbhost=DBHOST, dbport=DBPORT):
    return pymongo.connection.Connection(dbhost, dbport)

def get_db(dbhost=DBHOST, dbport=DBPORT, dbname=DBNAME):
    return pymongo.database.Database(get_connection(dbhost, dbport), dbname)

def get_dbdetails(db=None):
    '''return dbhostname, dbport, dbname'''
    if db is None:
        return DBHOST, DBPORT, DBNAME
    return db.connection.host, db.connection.port, db.name

def get_relist(relist):
    if type(relist) is not list:
        relist = relist.split(',')
    return [re.compile(item) for item in relist]

def get_whitelist(whitelist=None):
    if whitelist is None:
        return WHITELIST
    return get_relist(whitelist)


def get_blacklist(blacklist=None):
    if blacklist is None:
        return BLACKLIST
    return get_relist(blacklist)



#!/usr/bin/env python

from setuptools import setup
import glob, os

setup(name='moncov',
    version='0.5.9',
    description='Python remote code coverage using MongoDB',
    author='Vitaly Kuznetsov',
    author_email='vitty@redhat.com',
    url='https://github.com/RedHatQE/python-moncov',
    license="GPLv3+",
    provides=['moncov'],
    install_requires=['pymongo', 'aaargh', 'PyYAML', 'python-daemon', 'lxml'],
    packages=[
        'moncov',
        'moncov/stats'
        ],
    scripts=['bin/moncov'],
    data_files=[
        ('/etc', ['etc/moncov.yaml']),
        ('/usr/lib/systemd/system', ['lib/systemd/system/moncov.service'])
    ],
    classifiers=[
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
            'Programming Language :: Python',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Operating System :: POSIX',
            'Intended Audience :: Developers',
            'Development Status :: 4 - Beta'
    ]
)
